@@ -16,7 +16,7 @@
 // Copyright (C) 2006, 2008 Pino Toscano <pino@kde.org>
 // Copyright (C) 2007, 2010, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
 // Copyright (C) 2008 Hugo Mercier <hmercier31@gmail.com>
-// Copyright (C) 2008-2010, 2012-2014, 2016-2023 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2008-2010, 2012-2014, 2016-2024 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2009 Kovid Goyal <kovid@kovidgoyal.net>
 // Copyright (C) 2009 Ilya Gorenbein <igorenbein@finjan.com>
 // Copyright (C) 2012 Tobias Koening <tobias.koenig@kdab.com>
@@ -26,6 +26,7 @@
 // Copyright (C) 2019, 2020 Oliver Sander <oliver.sander@tu-dresden.de>
 // Copyright (C) 2020 Marek Kasik <mkasik@redhat.com>
 // Copyright (C) 2024 Pratham Gandhi <ppg.1382@gmail.com>
+// Copyright (C) 2025 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -36,6 +37,7 @@
 
 #include <cstddef>
 #include <cstring>
+#include <memory>
 #include "goo/gmem.h"
 #include "goo/GooString.h"
 #include "Error.h"
@@ -421,9 +423,9 @@ LinkGoTo::LinkGoTo(const Object *destObj)
 {
     // named destination
     if (destObj->isName()) {
-        namedDest = std::make_unique<GooString>(destObj->getName());
+        namedDest = std::make_unique<GooString>(destObj->getNameString());
     } else if (destObj->isString()) {
-        namedDest = std::unique_ptr<GooString>(destObj->getString()->copy());
+        namedDest = destObj->getString()->copy();
 
         // destination dictionary
     } else if (destObj->isArray()) {
@@ -449,14 +451,14 @@ LinkGoToR::LinkGoToR(Object *fileSpecObj, Object *destObj)
     // get file name
     Object obj1 = getFileSpecNameForPlatform(fileSpecObj);
     if (obj1.isString()) {
-        fileName = std::unique_ptr<GooString>(obj1.getString()->copy());
+        fileName = obj1.getString()->copy();
     }
 
     // named destination
     if (destObj->isName()) {
         namedDest = std::make_unique<GooString>(destObj->getName());
     } else if (destObj->isString()) {
-        namedDest = std::unique_ptr<GooString>(destObj->getString()->copy());
+        namedDest = destObj->getString()->copy();
 
         // destination dictionary
     } else if (destObj->isArray()) {
@@ -485,7 +487,7 @@ LinkLaunch::LinkLaunch(const Object *actionObj)
         if (!obj1.isNull()) {
             Object obj3 = getFileSpecNameForPlatform(&obj1);
             if (obj3.isString()) {
-                fileName = std::unique_ptr<GooString>(obj3.getString()->copy());
+                fileName = obj3.getString()->copy();
             }
         } else {
 #ifdef _WIN32
@@ -499,11 +501,11 @@ LinkLaunch::LinkLaunch(const Object *actionObj)
                 Object obj2 = obj1.dictLookup("F");
                 Object obj3 = getFileSpecNameForPlatform(&obj2);
                 if (obj3.isString()) {
-                    fileName = std::unique_ptr<GooString>(obj3.getString()->copy());
+                    fileName = obj3.getString()->copy();
                 }
                 obj2 = obj1.dictLookup("P");
                 if (obj2.isString()) {
-                    params = std::unique_ptr<GooString>(obj2.getString()->copy());
+                    params = obj2.getString()->copy();
                 }
             } else {
                 error(errSyntaxWarning, -1, "Bad launch-type link action");
@@ -535,7 +537,7 @@ LinkURI::LinkURI(const Object *uriObj, const std::optional<std::string> &baseURI
             // relative URI
             if (baseURI) {
                 uri = *baseURI;
-                if (uri.size() > 0) {
+                if (!uri.empty()) {
                     char c = uri.back();
                     if (c != '/' && c != '?') {
                         uri += '/';
@@ -565,7 +567,7 @@ LinkNamed::LinkNamed(const Object *nameObj)
 {
     hasNameFlag = false;
     if (nameObj->isName()) {
-        name = (nameObj->getName()) ? nameObj->getName() : "";
+        name = nameObj->getNameString();
         hasNameFlag = true;
     }
 }
@@ -719,7 +721,7 @@ LinkRendition::LinkRendition(const Object *obj)
                 operation = PlayRendition;
                 break;
             }
-        } else if (js == "") {
+        } else if (js.empty()) {
             error(errSyntaxWarning, -1, "Invalid Rendition action: no OP or JS field defined");
         }
     }
@@ -754,7 +756,7 @@ Object LinkJavaScript::createObject(XRef *xref, const std::string &js)
 {
     Dict *linkDict = new Dict(xref);
     linkDict->add("S", Object(objName, "JavaScript"));
-    linkDict->add("JS", Object(new GooString(js)));
+    linkDict->add("JS", Object(std::make_unique<GooString>(js)));
 
     return Object(linkDict);
 }
@@ -848,7 +850,7 @@ LinkResetForm::LinkResetForm(const Object *obj)
         for (int i = 0; i < obj1.arrayGetLength(); ++i) {
             const Object &obj2 = obj1.arrayGetNF(i);
             if (obj2.isName()) {
-                fields[i] = std::string(obj2.getName());
+                fields[i] = obj2.getNameString();
             } else if (obj2.isString()) {
                 fields[i] = obj2.getString()->toStr();
             } else if (obj2.isRef()) {
@@ -890,7 +892,7 @@ LinkSubmitForm::LinkSubmitForm(const Object *obj)
         for (int i = 0; i < objFields.arrayGetLength(); ++i) {
             const Object &objNF = objFields.arrayGetNF(i);
             if (objNF.isName()) {
-                fields[i] = std::string(objNF.getName());
+                fields[i] = objNF.getNameString();
             } else if (objNF.isString()) {
                 fields[i] = objNF.getString()->toStr();
             } else if (objNF.isRef()) {
@@ -944,20 +946,13 @@ Links::Links(Annots *annots)
         return;
     }
 
-    for (Annot *annot : annots->getAnnots()) {
+    for (const std::shared_ptr<Annot> &annot : annots->getAnnots()) {
 
         if (annot->getType() != Annot::typeLink) {
             continue;
         }
-
-        annot->incRefCnt();
-        links.push_back(static_cast<AnnotLink *>(annot));
+        links.push_back(std::static_pointer_cast<AnnotLink>(annot));
     }
 }
 
-Links::~Links()
-{
-    for (AnnotLink *link : links) {
-        link->decRefCnt();
-    }
-}
+Links::~Links() = default;
