@@ -1402,8 +1402,8 @@ void PDFDoc::writeObject(Object *obj, OutStream *outStr, XRef *xRef, unsigned in
         break;
     }
     case objName: {
-        GooString name(obj->getName());
-        outStr->printf("/%s ", sanitizedName(name.toStr()).c_str());
+        std::string name(obj->getNameString());
+        outStr->printf("/%s ", sanitizedName(name).c_str());
         break;
     }
     case objNull:
@@ -2244,7 +2244,7 @@ std::optional<PDFDoc::SignatureData> PDFDoc::createSignature(::Page *destPage, s
     field->setImageResource(imageResourceRef);
 
     Object refObj(ref);
-    AnnotWidget *signatureAnnot = new AnnotWidget(this, field->getObj(), &refObj, field.get());
+    auto signatureAnnot = std::make_shared<AnnotWidget>(this, field->getObj(), &refObj, field.get());
     signatureAnnot->setFlags(signatureAnnot->getFlags() | Annot::flagPrint | /*Annot::flagLocked | TODO */ Annot::flagNoRotate);
     Dict dummy(getXRef());
     auto appearCharacs = std::make_unique<AnnotAppearanceCharacs>(&dummy);
@@ -2256,9 +2256,6 @@ std::optional<PDFDoc::SignatureData> PDFDoc::createSignature(::Page *destPage, s
     border->setWidth(borderWidth);
     signatureAnnot->setBorder(std::move(border));
 
-    signatureAnnot->generateFieldAppearance();
-    signatureAnnot->updateAppearanceStream();
-
     FormWidget *formWidget = field->getWidget(field->getNumWidgets() - 1);
     formWidget->setWidgetAnnotation(signatureAnnot);
 
